# -*- coding: utf-8 -*-

class QueryObject(object):
    """
    Structure of css query.

    Use instance of this class with
    from_object method of QueryProcessor.
    """

    def __init__(self, tag, **selector):
        self.tag = tag
        self.selector = selector

    def __unicode__(self):
        return u'<{} {}>'.format(
            self.tag,
            u' '.join(
                [u'{}="{}"'.format(k, v) for k, v in self.selector.items()]
            ),
        )


class _Contains(object):
    """
    Marker for use contains inside css query
    """

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return self.value

    def __str__(self):
        return str(self.value)

    def __unicode__(self):
        return unicode(self.value)


contains = _Contains



# -*- coding: utf-8 -*-

import logging

from noseapp.utils.common import waiting_for
from noseapp.utils.common import TimeoutException
from selenium.common.exceptions import WebDriverException
from selenium.webdriver.remote.webelement import WebElement

from noseapp_selenium.query.exc import QueryError


logger = logging.getLogger(__name__)


DEFAULT_SLEEP = 0.01
DEFAULT_WAIT_TIMEOUT = 30


def _error_handler(e, client, css):
    """
    Extend error message
    """
    prefix = u' ' if e.message else u''

    if isinstance(client, WebElement):
        e.message += u'{}QueryProcessor(From: {}, CSS: {})\n\n--\nSEARCH AREA: {}\n--\n'.format(
            prefix,
            repr(client),
            css,
            client.get_attribute('innerHTML'),
        )
    else:
        e.message += u'{}QueryProcessor(From: {}, CSS: {})'.format(
            prefix,
            repr(client),
            css,
        )


def _execute(client, css, get_all=False, allow_polling=True):
    """
    Execute css query
    """
    logger.debug(u'CSS: {} Get all: {}'.format(css, 'Yes' if get_all else 'No'))

    css_executor = {
        True: 'find_elements_by_css_selector',
        False: 'find_element_by_css_selector',
    }

    try:
        if allow_polling:
            result = getattr(client, css_executor[get_all])(css)
        elif hasattr(client, 'disable_polling'):
            with client.disable_polling():
                result = getattr(client, css_executor[get_all])(css)
        else:
            result = getattr(client, css_executor[get_all])(css)

        return result

    except WebDriverException as e:
        _error_handler(e, client, css)
        raise
    except KeyError:
        raise QueryError('get_all param must be bool type only')


class QueryResult(object):
    """
    Execute actions by css query and returning result
    """

    def __init__(self, client, css):
        self._client = client
        self._css = css

    def __getattr__(self, item):
        return getattr(
            self._client.query.__class__(self.first()), item,
        )

    @property
    def exist(self):
        """
        Check element exist
        """
        self._client.config.implicitly_wait(0)

        try:
            el = _execute(self._client, self._css, allow_polling=False)

            if el:
                self._client.config.apply_implicitly_wait()
                return True
            self._client.config.apply_implicitly_wait()
            return False
        except WebDriverException:
            self._client.config.apply_implicitly_wait()
            return False
        except BaseException:
            self._client.config.apply_implicitly_wait()
            raise

    @property
    def with_wait(self):
        self.wait()
        return self

    def wait(self, timeout=None, sleep=None):
        """
        Waiting for web element exist
        """
        try:
            return waiting_for(
                lambda: self.exist,
                sleep=sleep or DEFAULT_SLEEP,
                timeout=timeout or DEFAULT_WAIT_TIMEOUT,
            )
        except TimeoutException:
            raise TimeoutException(
                'Could not wait web element with css "{}"'.format(self._css),
            )

    def get(self, index):
        """
        Get web element by index
        """
        try:
            return _execute(self._client, self._css, get_all=True)[index]
        except IndexError as e:
            e.message = 'Result does not have element with index "{}". Query: "{}".'.format(
                index, self._css,
            )
            raise

    def first(self):
        """
        Get first element on page
        """
        return _execute(self._client, self._css)

    def all(self):
        """
        Get all elements of appropriate query
        """
        return _execute(self._client, self._css, get_all=True)



# -*- coding: utf8 -*-

import time
from functools import wraps

from noseapp.utils.common import TimeoutException
from selenium.common.exceptions import WebDriverException


def make_object(web_element, allow_raise=True):
    """
    Convert web element to object.

    Example:

        input = driver.find_element...
        value = make_object(input).value
        css_value = make_object(input).css.background_image

    :type web_element: selenium.webdriver.remote.webdriver.WebElement
    """
    # TODO: __setattr__
    class WebElementToObject(object):

        def __init__(self, web_element, allow_raise):
            self.__web_element = web_element
            self.__allow_raise = allow_raise

        @property
        def css(self):
            return WebElementCssToObject(self.__web_element)

        def __getattr__(self, item):
            atr = self.__web_element.get_attribute(
                item.replace('_', '-'),
            )

            if atr:
                return atr

            if allow_raise:
                raise AttributeError('{} "{}"'.format(repr(self.__web_element), item))

    class WebElementCssToObject(object):

        def __init__(self, web_element):
            self.__web_element = web_element

        def __getattr__(self, item):
            return self.__web_element.value_of_css_property(
                item.replace('_', '-'),
            )

    return WebElementToObject(web_element, allow_raise)


def polling(callback=None, timeout=30, sleep=0.01):
    """
    Do sleep while wrapped function will be raised
    exception of WebDriverException class.

    Use timeout param for setting max seconds to waiting.
    This function will be used like decorator if callback is None.
    """
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            t_start = time.time()

            while time.time() <= t_start + timeout:
                try:
                    return f(*args, **kwargs)
                except WebDriverException:
                    time.sleep(sleep)
                    continue
            else:
                raise

        return wrapped

    if callback is not None:
        return wrapper(callback)

    return wrapper


class ReRaiseWebDriverException(BaseException):
    pass


def re_raise_wd_exc(callback=None, exc_cls=ReRaiseWebDriverException, message=None):
    """
    Except WebDriverException and re raising to custom exception.
    Also, TimeoutException will be excepting for re raise polling.
    """
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except (TimeoutException, WebDriverException) as e:
                raise exc_cls(
                    u"""
    Re raise web driver exception:

    * Message: {}
    * Original exception class: {}
    * Original message: {}
                    """.format(
                        message or '',
                        e.__class__.__name__,
                        e.message,
                    ),
                )

        return wrapped

    if callable(callback):
        return wrapper(callback)

    return wrapper


def get_query_from_driver(driver, wrapper=None):
    """
    Return QueryProcessor instance from driver.
    If wrapper is not None, at wrapper will be merged.

    :param driver: ProxyObject
    :param wrapper: QueryObject
    """
    if wrapper:
        return driver.query.from_object(
            wrapper,
        ).first().query

    return driver.query



# -*- coding: utf-8 -*-

import logging

from noseapp_selenium.query.base import contains
from noseapp_selenium.query.result import QueryResult


logger = logging.getLogger(__name__)


REPLACE_ATTRIBUTES = {
    '_id': 'id',
    '_class': 'class',
    '_type': 'type',
}

REPLACE_TAGS = {
    'link': 'a',
}


def _replace_attribute(atr_name):
    """
    Replacing attribute name for
    excluding conflict with names of globals
    """
    return REPLACE_ATTRIBUTES.get(atr_name, atr_name).replace('_', '-')


def _replace_tag(tag_name):
    """
    Replacing tag name for usability
    """
    return REPLACE_TAGS.get(tag_name, tag_name)


def make_result(client, tag):
    """
    :type client: selenium.webdriver.remote.webdriver.WebDriver
    :param tag: html tag name
    """
    def handle(**selector):
        query = [_replace_tag(tag)]

        def get_format(value):
            if isinstance(value, contains):
                return u'[{}*="{}"]'
            return u'[{}="{}"]'

        query.extend(
            (
                get_format(val).format(_replace_attribute(atr), val)
                for atr, val in selector.items()
            ),
        )

        return QueryResult(client, ''.join(query))

    return handle



# -*- coding: utf-8 -*-

from noseapp_selenium.query.base import contains
from noseapp_selenium.query.base import QueryObject
from noseapp_selenium.query.result import QueryResult
from noseapp_selenium.query.processor import QueryProcessor


__all__ = (
    contains,
    QueryObject,
    QueryProcessor,
)



# -*- coding: utf-8 -*-

from selenium.webdriver.remote.webelement import WebElement

from noseapp_selenium.query.base import QueryObject
from noseapp_selenium.query.handler import make_result


class QueryProcessor(object):
    """
    Example:

        driver = WebDriver()
        query = QueryProcessor(driver)

        search_wrapper = query.div(id='search').first()
        search_field = query(search_wrapper).input(_class='search').first()
        search_field.send_keys(*'Hello World!')
    """

    def __init__(self, client):
        self._client = client

    def __getattr__(self, item):
        return make_result(self._client, item)

    def __call__(self, client):
        return self.__class__(client)

    @property
    def client(self):
        return self._client

    def from_object(self, obj):
        """
        Creating result from object

        :type obj: QueryObject
        """
        if not isinstance(obj, QueryObject):
            raise TypeError('"{}" is not QueryObject instance'.format(type(obj)))

        return self.__getattr__(obj.tag)(**obj.selector)

    def get_text(self):
        """
        Get text from page or web element
        """
        if isinstance(self._client.orig(), WebElement):
            return self._client.text

        return self._client.find_element_by_tag_name('body').text
